with Controller;
with Results;
with Ada.Text_IO;

package body Primes is

   type Index is range 0 .. Common.Search_Limit;

   --  STATIC VARIABLES
   ---------------------------------
   --  Sifter records the prime numbers known to this partition
   Sifter  : array (Index) of Natural;

   --  Tail is the index in Sifter of the first prime found by the Head partition
   --  and a sentinel for Sieve to tell its first invocation from all others
   Tail    : Index := 0;

   --  Head points to the index in Sifter where to begin testing againt Number
   --  (we don't want to divide by 0)
   Head : Index := 1;

   --  Candidate records which number we are testing so as to reset Head
   --  when a new number is to be tested against the values in Sifter
   Candidate     : Natural := 1;

   --  attribute 'Partition_ID of unit U returns an integer
   --  which is the unique id that identifies the partition
   --  on which unit U is located
   Self : Common.Partition_ID := Primes'Partition_ID;

   --  distributed objects for the current and next partition
   --  (Pool_Access is a class-wide type)
   Own_Pool  : aliased My_Pool_Type;

   My_Neighbour    : Controller.Pool_Access := null;
   ---------------------------------
   --  END STATIC VARIABLES

   procedure Reset (Number: in Natural) is
      --  to gain visibility on the operator on Pool_Type
      use Controller;
   begin
      --  we reset Head to 1 in case we received a new Number to test
      --  (recall that initially Head := 2 and Candidate :=1)
      if Candidate /= Number then
         Candidate := Number; --  Candidate is a static variable
         Head := 1;
      end if;
   end Reset;

   -----------------------
   procedure Slow_Down (Length : Positive) is
   begin
      for Num in 0 .. Length loop
         for Num2 in 0 .. 4 loop
            null;
         end loop;
      end loop;
   end Slow_Down;
   -----------------------

   protected Locker is
      procedure Finder (Number : in Natural);
   end Locker;

   protected body Locker is

      procedure Finder (Number : in Natural) is
      begin
         -- we count the calls to Sieve across all partitions
         Results.Increment;

         --  if this is a "new" call then we reset Counter
         Reset (Number);

         --  (Head > Tail) is the sentinel that tells whether Number is a prime:
         --  this condition holds by design at the first pass in Sieve
         --  and also when our own call to the neighbor's Sieve returns here
         --  without having found a divisor
         if Head <= Tail then
            --  do we have a divisor for Number in Sifter?
            if (Number mod Sifter (Head)) = 0 then
               --  we found a divisor hence we record it in Store
               Results.Store (Sifter (Head), Self);
            else
               --  we did not find a divisor at position Head in Sifter:
               --  if the call comes back here we need to look next
               Head := Head + 1;

               --  we make an RMI to our neighbor partition
               --  to ask it to continue checking Number against its own Sifter
               Sieve (My_Neighbour, Number);

               -- ++++++++++++++++
               --  what happens if we defer the increment to Head when we
               --  allow multiple calls to Sieve to interleave in the same partition?
               Slow_Down (1_000_000);
               --  +++++++++++++++
            end if;
         else
            --  Number is a prime
            --  hence we must add it to Sifter and record it in Store
            Tail := Tail + 1;
            Sifter (Tail) := Number;
            Results.Store (Sifter (Tail), Self);
         end if;
      end Finder;
   end Locker;

   procedure Sieve
     (Pool     : access My_Pool_Type;
      Number   : in Natural) is
   begin
      Locker.Finder (Number);
   end Sieve;

begin

   --  in order that mutiple instances of this unit may be identified
   --  by the Main unit that invokes procedure Sieve we register
   --  the distributed object Own_Pool that belongs to the current partition
   --  to the care of the Controller unit.
   --  We do so by using a dynamically bound call dispatched according
   --  to the value of its controlling parameter (Own_Pool'access)
   --  which is of an access-to-class-wide type
   --  (this statement is executed during the elaboration of the partition)

   if not Controller.Register (Own_Pool'Access, Self) then
      Ada.Text_IO.Put_Line ("Can't register partition: fatal error.");
      raise Program_Error;
   end if;

   --  now that everything is well we inquire our neighbour in the ring
   My_Neighbour := Controller.Next (Self);

end Primes;
